@@ -4,6 +4,7 @@ from functools import lru_cache
 from typing import Any
 from typing import Dict
 from typing import List
+from typing import Optional
 from typing import Set
 from typing import Tuple
 from typing import Union
@@ -11,15 +12,23 @@ from typing import Union
 import py
 
 import _pytest._code
+from _pytest._code.code import ExceptionChainRepr
+from _pytest._code.code import ExceptionInfo
+from _pytest._code.code import ReprExceptionInfo
+from _pytest.compat import cached_property
 from _pytest.compat import getfslineno
+from _pytest.config import Config
+from _pytest.fixtures import FixtureDef
+from _pytest.fixtures import FixtureLookupError
+from _pytest.fixtures import FixtureLookupErrorRepr
 from _pytest.mark.structures import Mark
 from _pytest.mark.structures import MarkDecorator
 from _pytest.mark.structures import NodeKeywords
-from _pytest.outcomes import fail
+from _pytest.outcomes import Failed
 
 if False:  # TYPE_CHECKING
     # Imported here due to circular import.
-    from _pytest.fixtures import FixtureDef
+    from _pytest.main import Session  # noqa: F401
 
 SEP = "/"
 
@@ -69,8 +78,14 @@ class Node:
     Collector subclasses have children, Items are terminal nodes."""
 
     def __init__(
-        self, name, parent=None, config=None, session=None, fspath=None, nodeid=None
-    ):
+        self,
+        name,
+        parent: Optional["Node"] = None,
+        config: Optional[Config] = None,
+        session: Optional["Session"] = None,
+        fspath: Optional[py.path.local] = None,
+        nodeid: Optional[str] = None,
+    ) -> None:
         #: a unique name within the scope of the parent node
         self.name = name
 
@@ -78,10 +93,20 @@ class Node:
         self.parent = parent
 
         #: the pytest config object
-        self.config = config or parent.config
+        if config:
+            self.config = config
+        else:
+            if not parent:
+                raise TypeError("config or parent must be provided")
+            self.config = parent.config
 
         #: the session this node is part of
-        self.session = session or parent.session
+        if session:
+            self.session = session
+        else:
+            if not parent:
+                raise TypeError("session or parent must be provided")
+            self.session = parent.session
 
         #: filesystem path where this node was collected from (can be None)
         self.fspath = fspath or getattr(parent, "fspath", None)
@@ -102,6 +127,8 @@ class Node:
             assert "::()" not in nodeid
             self._nodeid = nodeid
         else:
+            if not self.parent:
+                raise TypeError("nodeid or parent must be provided")
             self._nodeid = self.parent.nodeid
             if self.name != "()":
                 self._nodeid += "::" + self.name
@@ -139,8 +166,7 @@ class Node:
                 )
             )
         path, lineno = get_fslocation_from_item(self)
-        # Type ignored: https://github.com/python/typeshed/pull/3121
-        warnings.warn_explicit(  # type: ignore
+        warnings.warn_explicit(
             warning,
             category=None,
             filename=str(path),
@@ -166,7 +192,7 @@ class Node:
         """ return list of all parent collectors up to self,
             starting from root of collection tree. """
         chain = []
-        item = self
+        item = self  # type: Optional[Node]
         while item is not None:
             chain.append(item)
             item = item.parent
@@ -247,7 +273,7 @@ class Node:
     def getparent(self, cls):
         """ get the next parent node (including ourself)
         which is an instance of the given class"""
-        current = self
+        current = self  # type: Optional[Node]
         while current and not isinstance(current, cls):
             current = current.parent
         return current
@@ -255,13 +281,13 @@ class Node:
     def _prunetraceback(self, excinfo):
         pass
 
-    def _repr_failure_py(self, excinfo, style=None):
-        # Type ignored: see comment where fail.Exception is defined.
-        if excinfo.errisinstance(fail.Exception):  # type: ignore
+    def _repr_failure_py(
+        self, excinfo: ExceptionInfo[Union[Failed, FixtureLookupError]], style=None
+    ) -> Union[str, ReprExceptionInfo, ExceptionChainRepr, FixtureLookupErrorRepr]:
+        if isinstance(excinfo.value, Failed):
             if not excinfo.value.pytrace:
                 return str(excinfo.value)
-        fm = self.session._fixturemanager
-        if excinfo.errisinstance(fm.FixtureLookupError):
+        if isinstance(excinfo.value, FixtureLookupError):
             return excinfo.value.formatrepr()
         if self.config.getoption("fulltrace", False):
             style = "long"
@@ -299,7 +325,9 @@ class Node:
             truncate_locals=truncate_locals,
         )
 
-    def repr_failure(self, excinfo, style=None):
+    def repr_failure(
+        self, excinfo, style=None
+    ) -> Union[str, ReprExceptionInfo, ExceptionChainRepr, FixtureLookupErrorRepr]:
         return self._repr_failure_py(excinfo, style)
 
 
@@ -365,8 +393,9 @@ def _check_initialpaths_for_relpath(session, fspath):
 
 
 class FSCollector(Collector):
-    def __init__(self, fspath, parent=None, config=None, session=None, nodeid=None):
-        fspath = py.path.local(fspath)  # xxx only for test_resultlog.py?
+    def __init__(
+        self, fspath: py.path.local, parent=None, config=None, session=None, nodeid=None
+    ) -> None:
         name = fspath.basename
         if parent is not None:
             rel = fspath.relto(parent.fspath)
@@ -426,16 +455,12 @@ class Item(Node):
         if content:
             self._report_sections.append((when, key, content))
 
-    def reportinfo(self):
+    def reportinfo(self) -> Tuple[str, Optional[int], str]:
         return self.fspath, None, ""
 
-    @property
-    def location(self):
-        try:
-            return self._location
-        except AttributeError:
-            location = self.reportinfo()
-            fspath = self.session._node_location_to_relpath(location[0])
-            location = (fspath, location[1], str(location[2]))
-            self._location = location
-            return location
+    @cached_property
+    def location(self) -> Tuple[str, Optional[int], str]:
+        location = self.reportinfo()
+        fspath = self.session._node_location_to_relpath(location[0])
+        assert type(location[2]) is str
+        return (fspath, location[1], location[2])
